@@ -69,9 +69,13 @@ class SqliteQueryCompiler(QueryCompiler):
             statement += ' IF NOT EXISTS'
         clause.nodes[0] = SQL(statement)  # Overwrite the statement.
 
+        table_options = getattr(model_class._meta, 'options', None) or {}
         if options:
+            table_options.update(options)
+
+        if table_options:
             columns_constraints = clause.nodes[-1]
-            for k, v in options.items():
+            for k, v in sorted(table_options.items()):
                 if isinstance(v, Field):
                     value = v._as_entity(with_table=True)
                 elif inspect.isclass(v) and issubclass(v, Model):
@@ -166,6 +170,82 @@ class FTSModel(VirtualModel):
                 .where(cls.match(term))
                 .order_by(SQL(alias).desc()))
 
+class _MockFieldMixin(object):
+    def add_to_class(self, model_class, name):
+        super(_MockFieldMixin, self).add_to_class(model_class, name)
+        del model_class._meta.fields[self.name]
+        del model_class._meta.columns[self.db_column]
+
+class _MockIntegerField(_MockFieldMixin, IntegerField):
+    pass
+
+class _MockCharField(_MockFieldMixin, CharField):
+    pass
+
+def ClosureTable(model_class, foreign_key=None):
+    """Model factory for the transitive closure extension."""
+    if foreign_key is None:
+        for field_obj in model_class._meta.rel.values():
+            if field_obj.rel_model is model_class:
+                foreign_key = field_obj
+                break
+        else:
+            raise ValueError('Unable to find self-referential foreign key.')
+    primary_key = model_class._meta.primary_key
+
+    class BaseClosureTable(VirtualModel):
+        _extension = 'transitive_closure'
+
+        depth = _MockIntegerField()
+        id = _MockIntegerField()
+        idcolumn = _MockIntegerField()
+        parentcolumn = _MockIntegerField()
+        root = _MockIntegerField()
+        tablename = _MockCharField()
+
+        @classmethod
+        def descendants(cls, node, depth=None, include_node=False):
+            query = (model_class
+                     .select()
+                     .join(cls, on=(primary_key == cls.id))
+                     .where(cls.root == node))
+            if depth is not None:
+                query = query.where(cls.depth == depth)
+            elif not include_node:
+                query = query.where(cls.depth > 0)
+            return query
+
+        @classmethod
+        def ancestors(cls, node, depth=None, include_node=False):
+            query = (model_class
+                     .select()
+                     .join(cls, on=(primary_key == cls.root))
+                     .where(cls.id == node))
+            if depth:
+                query = query.where(cls.depth == depth)
+            elif not include_node:
+                query = query.where(cls.depth > 0)
+            return query
+
+        @classmethod
+        def siblings(cls, node, include_node=False):
+            fk_value = node._data.get(foreign_key.name)
+            query = model_class.select().where(foreign_key == fk_value)
+            if not include_node:
+                query = query.where(primary_key != node)
+            return query
+
+    class Meta:
+        database = model_class._meta.database
+        options = {
+            'tablename': model_class._meta.db_table,
+            'idcolumn': model_class._meta.primary_key.db_column,
+            'parentcolumn': foreign_key.db_column}
+        primary_key = False
+
+    name = '%sClosure' % model_class.__name__
+    return type(name, (BaseClosureTable,), {'Meta': Meta})
+
 
 class SqliteExtDatabase(SqliteDatabase):
     """
@@ -182,6 +262,7 @@ class SqliteExtDatabase(SqliteDatabase):
         self._aggregates = {}
         self._collations = {}
         self._functions = {}
+        self._extensions = set([])
         self._row_factory = None
         self.register_function(rank, 'rank', 1)
         self.register_function(bm25, 'bm25', -1)
@@ -196,6 +277,10 @@ class SqliteExtDatabase(SqliteDatabase):
             conn.create_function(name, num_params, fn)
         if self._row_factory:
             conn.row_factory = self._row_factory
+        if self._extensions:
+            conn.enable_load_extension(True)
+            for extension in self._extensions:
+                conn.load_extension(extension)
         return conn
 
     def _argc(self, fn):
@@ -235,6 +320,9 @@ class SqliteExtDatabase(SqliteDatabase):
             return fn
         return decorator
 
+    def load_extension(self, extension):
+        self._extensions.add(extension)
+
     def unregister_aggregate(self, name):
         del(self._aggregates[name])
 
@@ -244,6 +332,9 @@ class SqliteExtDatabase(SqliteDatabase):
     def unregister_function(self, name):
         del(self._functions[name])
 
+    def unload_extension(self, extension):
+        self._extensions.remove(extension)
+
     def row_factory(self, fn):
         self._row_factory = fn
 
