@@ -1,13 +1,22 @@
-import sqlite3; sqlite3.enable_callback_tracebacks(True)
+import os
+import sqlite3
+try:
+    sqlite3.enable_callback_tracebacks(True)
+except AttributeError:
+    pass
 import unittest
 
 from peewee import *
-from playhouse import sqlite_ext as sqe
+from peewee import print_
+from playhouse.sqlite_ext import *
 
 # use a disk-backed db since memory dbs only exist for a single connection and
 # we need to share the db w/2 for the locking tests.  additionally, set the
 # sqlite_busy_timeout to 100ms so when we test locking it doesn't take forever
-ext_db = sqe.SqliteExtDatabase('tmp.db', timeout=.1)
+ext_db = SqliteExtDatabase('tmp.db', timeout=.1)
+
+CLOSURE_EXTENSION = os.environ.get('CLOSURE_EXTENSION')
+TEST_VERBOSITY = int(os.environ.get('PEEWEE_TEST_VERBOSITY') or 1)
 
 # test aggregate.
 class WeightedAverage(object):
@@ -55,29 +64,29 @@ ext_db.register_collation(collate_reverse)
 ext_db.register_function(title_case)
 
 
-class BaseExtModel(sqe.Model):
+class BaseExtModel(Model):
     class Meta:
         database = ext_db
 
 class Post(BaseExtModel):
     message = TextField()
 
-class FTSPost(Post, sqe.FTSModel):
+class FTSPost(Post, FTSModel):
     """Automatically managed and populated via the Post model."""
     pass
 
-class FTSDoc(sqe.FTSModel):
+class FTSDoc(FTSModel):
     """Manually managed and populated using queries."""
     message = TextField()
     class Meta:
         database = ext_db
 
-class ManagedDoc(sqe.FTSModel):
+class ManagedDoc(FTSModel):
     message = TextField()
     class Meta:
         database = ext_db
 
-class MultiColumn(sqe.FTSModel):
+class MultiColumn(FTSModel):
     c1 = CharField(default='')
     c2 = CharField(default='')
     c3 = CharField(default='')
@@ -92,6 +101,18 @@ class Values(BaseExtModel):
     weight = FloatField()
 
 
+class TestVirtualModel(VirtualModel):
+    _extension = 'test_ext'
+    class Meta:
+        database = ext_db
+        options = {
+            'foo': 'bar',
+            'baze': 'nugget'}
+        primary_key = False
+
+class TestVirtualModelChild(TestVirtualModel):
+    pass
+
 
 class SqliteExtTestCase(unittest.TestCase):
     messages = [
@@ -123,9 +144,32 @@ class SqliteExtTestCase(unittest.TestCase):
         ManagedDoc.create_table(tokenize='porter', content=Post.message)
         FTSDoc.create_table(tokenize='porter')
 
+    def test_virtual_model_options(self):
+        compiler = ext_db.compiler()
+        sql, params = compiler.create_table(TestVirtualModel)
+        self.assertEqual(sql, (
+            'CREATE VIRTUAL TABLE "testvirtualmodel" USING test_ext '
+            '(baze=nugget, foo=bar)'))
+        self.assertEqual(params, [])
+
+        sql, params = compiler.create_table(TestVirtualModelChild)
+        self.assertEqual(sql, (
+            'CREATE VIRTUAL TABLE "testvirtualmodelchild" USING test_ext '
+            '("id" INTEGER NOT NULL PRIMARY KEY, baze=nugget, foo=bar)'))
+        self.assertEqual(params, [])
+
+        test_options = {'baze': 'nugz', 'huey': 'mickey'}
+        sql, params = compiler.create_table(
+            TestVirtualModel,
+            options=test_options)
+        self.assertEqual(sql, (
+            'CREATE VIRTUAL TABLE "testvirtualmodel" USING test_ext '
+            '(baze=nugz, foo=bar, huey=mickey)'))
+        self.assertEqual(params, [])
+
     def test_pk_autoincrement(self):
         class AutoInc(Model):
-            id = sqe.PrimaryKeyAutoIncrementField()
+            id = PrimaryKeyAutoIncrementField()
             foo = CharField()
 
         compiler = ext_db.compiler()
@@ -298,7 +342,7 @@ class SqliteExtTestCase(unittest.TestCase):
         ])
 
         pq = (ModelClass
-              .select(sqe.Rank(ModelClass))
+              .select(Rank(ModelClass))
               .where(ModelClass.match('faithful'))
               .tuples())
         self.assertEqual([x[0] for x in pq], [.2] * 5)
@@ -441,3 +485,212 @@ class SqliteExtTestCase(unittest.TestCase):
         res = conn2.execute('select message from post order by message;')
         self.assertEqual([x[0] for x in res.fetchall()], [
             'p2', 'p2', 'p4'])
+
+
+class TestTransitiveClosure(unittest.TestCase):
+    def test_model_factory(self):
+        class Category(BaseExtModel):
+            name = CharField()
+            parent = ForeignKeyField('self', null=True)
+
+        Closure = ClosureTable(Category)
+        self.assertEqual(Closure._extension, 'transitive_closure')
+        self.assertEqual(Closure._meta.columns, {})
+        self.assertEqual(Closure._meta.fields, {})
+        self.assertFalse(Closure._meta.primary_key)
+        self.assertEqual(Closure._meta.options, {
+            'idcolumn': 'id',
+            'parentcolumn': 'parent_id',
+            'tablename': 'category',
+        })
+
+        class Alt(BaseExtModel):
+            pk = PrimaryKeyField()
+            ref = ForeignKeyField('self', null=True)
+
+        Closure = ClosureTable(Alt)
+        self.assertEqual(Closure._meta.columns, {})
+        self.assertEqual(Closure._meta.fields, {})
+        self.assertFalse(Closure._meta.primary_key)
+        self.assertEqual(Closure._meta.options, {
+            'idcolumn': 'pk',
+            'parentcolumn': 'ref_id',
+            'tablename': 'alt',
+        })
+
+        class NoForeignKey(BaseExtModel):
+            pass
+        self.assertRaises(ValueError, ClosureTable, NoForeignKey)
+
+if CLOSURE_EXTENSION:
+    class TestTransitiveClosureIntegration(unittest.TestCase):
+        tree = {
+            'books': [
+                {'fiction': [
+                    {'scifi': [
+                        {'hard scifi': []},
+                        {'dystopian': []}]},
+                    {'westerns': []},
+                    {'classics': []},
+                ]},
+                {'non-fiction': [
+                    {'biographies': []},
+                    {'essays': []},
+                ]},
+            ]
+        }
+
+        def setUp(self):
+            ext_db.load_extension(CLOSURE_EXTENSION.rstrip('.so'))
+            ext_db.close()
+
+        def initialize_models(self):
+            class Category(BaseExtModel):
+                name = CharField()
+                parent = ForeignKeyField('self', null=True)
+                @classmethod
+                def g(cls, name):
+                    return cls.get(cls.name == name)
+
+            Closure = ClosureTable(Category)
+            ext_db.drop_tables([Category, Closure], True)
+            ext_db.create_tables([Category, Closure])
+
+            def build_tree(nodes, parent=None):
+                for name, subnodes in nodes.items():
+                    category = Category.create(name=name, parent=parent)
+                    if subnodes:
+                        for subnode in subnodes:
+                            build_tree(subnode, category)
+
+            build_tree(self.tree)
+            return Category, Closure
+
+        def assertNodes(self, query, *expected):
+            self.assertEqual(
+                set([category.name for category in query]),
+                set(expected))
+
+        def test_build_tree(self):
+            Category, Closure = self.initialize_models()
+            self.assertEqual(Category.select().count(), 10)
+
+        def test_descendants(self):
+            Category, Closure = self.initialize_models()
+            books = Category.g('books')
+            self.assertNodes(
+                Closure.descendants(books),
+                'fiction', 'scifi', 'hard scifi', 'dystopian',
+                'westerns', 'classics', 'non-fiction', 'biographies', 'essays')
+
+            self.assertNodes(Closure.descendants(books, 0), 'books')
+            self.assertNodes(
+                Closure.descendants(books, 1), 'fiction', 'non-fiction')
+            self.assertNodes(
+                Closure.descendants(books, 2),
+                'scifi', 'westerns', 'classics', 'biographies', 'essays')
+            self.assertNodes(
+                Closure.descendants(books, 3), 'hard scifi', 'dystopian')
+
+            fiction = Category.g('fiction')
+            self.assertNodes(
+                Closure.descendants(fiction),
+                'scifi', 'hard scifi', 'dystopian', 'westerns', 'classics')
+            self.assertNodes(
+                Closure.descendants(fiction, 1),
+                'scifi', 'westerns', 'classics')
+            self.assertNodes(
+                Closure.descendants(fiction, 2), 'hard scifi', 'dystopian')
+
+            self.assertNodes(
+                Closure.descendants(Category.g('scifi')),
+                'hard scifi', 'dystopian')
+            self.assertNodes(
+                Closure.descendants(Category.g('scifi'), include_node=True),
+                'scifi', 'hard scifi', 'dystopian')
+            self.assertNodes(Closure.descendants(Category.g('hard scifi'), 1))
+
+        def test_ancestors(self):
+            Category, Closure = self.initialize_models()
+
+            hard_scifi = Category.g('hard scifi')
+            self.assertNodes(
+                Closure.ancestors(hard_scifi),
+                'scifi', 'fiction', 'books')
+            self.assertNodes(
+                Closure.ancestors(hard_scifi, include_node=True),
+                'hard scifi', 'scifi', 'fiction', 'books')
+            self.assertNodes(Closure.ancestors(hard_scifi, 2), 'fiction')
+            self.assertNodes(Closure.ancestors(hard_scifi, 3), 'books')
+
+            non_fiction = Category.g('non-fiction')
+            self.assertNodes(Closure.ancestors(non_fiction), 'books')
+            self.assertNodes(Closure.ancestors(non_fiction, include_node=True),
+                             'non-fiction', 'books')
+            self.assertNodes(Closure.ancestors(non_fiction, 1), 'books')
+
+            books = Category.g('books')
+            self.assertNodes(Closure.ancestors(books, include_node=True),
+                             'books')
+            self.assertNodes(Closure.ancestors(books))
+            self.assertNodes(Closure.ancestors(books, 1))
+
+        def test_siblings(self):
+            Category, Closure = self.initialize_models()
+
+            self.assertNodes(
+                Closure.siblings(Category.g('hard scifi')), 'dystopian')
+            self.assertNodes(
+                Closure.siblings(Category.g('hard scifi'), include_node=True),
+                'hard scifi', 'dystopian')
+            self.assertNodes(
+                Closure.siblings(Category.g('classics')), 'scifi', 'westerns')
+            self.assertNodes(
+                Closure.siblings(Category.g('classics'), include_node=True),
+                'scifi', 'westerns', 'classics')
+            self.assertNodes(
+                Closure.siblings(Category.g('fiction')), 'non-fiction')
+
+        def test_tree_changes(self):
+            Category, Closure = self.initialize_models()
+            books = Category.g('books')
+            fiction = Category.g('fiction')
+            dystopian = Category.g('dystopian')
+            essays = Category.g('essays')
+            new_root = Category.create(name='products')
+            Category.create(name='magazines', parent=new_root)
+            books.parent = new_root
+            books.save()
+            dystopian.delete_instance()
+            essays.parent = books
+            essays.save()
+            Category.create(name='rants', parent=essays)
+            Category.create(name='poetry', parent=books)
+
+            query = (Category
+                     .select(Category.name, Closure.depth)
+                     .join(Closure, on=(Category.id == Closure.id))
+                     .where(Closure.root == new_root)
+                     .order_by(Closure.depth, Category.name)
+                     .tuples())
+            self.assertEqual(list(query), [
+                ('products', 0),
+                ('books', 1),
+                ('magazines', 1),
+                ('essays', 2),
+                ('fiction', 2),
+                ('non-fiction', 2),
+                ('poetry', 2),
+                ('biographies', 3),
+                ('classics', 3),
+                ('rants', 3),
+                ('scifi', 3),
+                ('westerns', 3),
+                ('hard scifi', 4),
+            ])
+
+        def tearDown(self):
+            ext_db.unload_extension(CLOSURE_EXTENSION.rstrip('.so'))
+
+elif TEST_VERBOSITY > 0:
+    print_('Skipping transitive closure integration tests.')
